@@ -1,6 +1,7 @@
+import shutil
 from functools import partial
-from pathlib import Path
 from math import ceil
+from pathlib import Path
 
 import PySide6.QtCore as Qc
 import PySide6.QtGui as Qg
@@ -18,8 +19,8 @@ from deepqt import __program__, __version__
 from deepqt import helpers as hp
 from deepqt.driver_api_config import ConfigureAccount
 from deepqt.file_table import Column, make_output_filename
-from deepqt.ui_generated_files.ui_mainwindow import Ui_MainWindow
 from deepqt.helpers import show_warning, show_info, show_question
+from deepqt.ui_generated_files.ui_mainwindow import Ui_MainWindow
 
 
 # noinspection PyUnresolvedReferences
@@ -42,6 +43,7 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         self.setWindowIcon(Qg.QIcon(":/icons/logo.png"))
         self.translating = False  # If true, the translation is in progress.
         self.glossary = st.Glossary()  # Create a dummy glossary
+        nuke_epub_cache()
 
         self.initialize_ui()
 
@@ -127,6 +129,8 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         self.abort_translation_worker.emit()
         if self.threadpool.activeThreadCount():
             self.threadpool.waitForDone()
+
+        nuke_epub_cache()
         event.accept()
 
     def connect_combobox_slots(self):
@@ -251,6 +255,7 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         self.config.fixed_output_path = self.lineEdit_file_out_dir.text()
         if save:
             self.config.save()
+        self.text_output_changed.emit()
 
     def use_glossary_toggled(self):
         """
@@ -398,8 +403,8 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         """
         self.glossary = glossary
         terms_found = len(glossary)
-        self.statusbar.showMessage(f"Loaded {terms_found} {hp.f_plural(terms_found, 'term', 'terms')} from glossary.")
-        logger.info(f"Glossary loaded, {terms_found} {hp.f_plural(terms_found, 'term', 'terms')} found.")
+        self.statusbar.showMessage(f"Loaded {terms_found} {hp.f_plural(terms_found, 'term')} from glossary.")
+        logger.info(f"Glossary loaded, {terms_found} {hp.f_plural(terms_found, 'term')} found.")
         logger.debug(f"Glossary dump: \n{glossary}")
         self.text_params_changed.emit(self.glossary)
 
@@ -460,9 +465,9 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         remaining_chars = allowed_chars - total_chars
         warning_msg = (
             f"You are about to translate {hp.format_char_count(total_chars)} "
-            f"{hp.f_plural(total_chars, 'character', 'characters')},\n"
+            f"{hp.f_plural(total_chars, 'character')},\n"
             f"leaving you with {hp.format_char_count(remaining_chars)} "
-            f"{hp.f_plural(remaining_chars, 'character', 'characters')}."
+            f"{hp.f_plural(remaining_chars, 'character')}."
             f"\nProceed?"
         )
         if api_usage.character.limit_reached:
@@ -470,7 +475,7 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         elif total_chars > allowed_chars:
             warning_msg = (
                 f"You are about to translate {hp.format_char_count(total_chars)} "
-                f"{hp.f_plural(total_chars, 'character', 'characters')}, "
+                f"{hp.f_plural(total_chars, 'character')}, "
                 f"which exceeds your character limit of {hp.format_char_count(allowed_chars)}.\nProceed anyway?"
             )
         # Ask the user if he wants to proceed, just in case.
@@ -546,6 +551,22 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         self.statusbar.showMessage("Aborting translation...")
 
     def write_output_file(self, file_id: str):
+        """
+        Write the output file to disk.
+        Check the type of the file (text or epub) and run the appropriate function.
+
+        :param file_id: InputFile ID.
+        """
+
+        file = self.file_table.files[file_id]
+        if isinstance(file, st.TextFile):
+            self.write_output_text_file(file_id)
+        elif isinstance(file, st.EpubFile):
+            self.write_output_epub_file(file_id)
+        else:
+            raise TypeError(f"Unknown file type: {file}")
+
+    def write_output_text_file(self, file_id: str):
         """
         Write the translation of a file to the assigned output path.
 
@@ -584,6 +605,31 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
                 show_warning(self, "Output Error", f"Failed to write translation to {path_out}.")
                 self.file_table.show_file_progress(file_id, "Could not write output!")
 
+    def write_output_epub_file(self, file_id: str):
+        """
+        Write the translation of a file to the assigned output path.
+
+        Check the status of the file. We decide 3 cases:
+        1. File is not translated.
+           - Skip this file.
+        2. File is translated.
+           - Write the translation to the output path.
+        3. File is partially translated.
+           - Dump the translation to the output path, if so configured.
+
+        :param file_id: InputFile ID.
+        """
+        file: st.EpubFile = self.file_table.files[file_id]
+        path_out = make_output_filename(file, self.config)
+
+        if not file.translation_incomplete() and not file.is_translated():
+            # Skip this because we have nothing to dump.
+            logger.info(f"Skipping file {file_id} because nothing has been translated.")
+            self.file_table.show_file_progress(file_id, "Not translated.")
+            return
+
+        file.write(st.ProcessLevel.TRANSLATED, path_out)
+
     """
     Log file
     """
@@ -723,7 +769,7 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
             return
 
         time_total = ceil(self.config.avg_time_per_mille * char_total / 1000)
-        char_text = hp.format_char_count(char_total) + hp.f_plural(char_total, " character", " characters")
+        char_text = hp.format_char_count(char_total) + hp.f_plural(char_total, " character")
         time_text = f"Approx. {hp.f_time(time_total)}"
         self.label_stats.setText(char_text + "   " + time_text)
 
@@ -737,7 +783,7 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
         char_text = hp.format_char_count(processed_chars) + " / " + hp.format_char_count(char_total)
         time_total = ceil(self.config.avg_time_per_mille * (char_total - processed_chars) / 1000)
         self.label_progress.setText(
-            f"Translated {char_text} {hp.f_plural(char_total, 'character', 'characters')}\n"
+            f"Translated {char_text} {hp.f_plural(char_total, 'character')}\n"
             f"Approximately {hp.f_time(time_total)} remaining"
         )
         self.progressBar.setValue(processed_chars / char_total * 100)
@@ -769,3 +815,23 @@ class MainWindow(Qw.QMainWindow, Ui_MainWindow):
                     </body>
                 </html>""",
         )
+
+
+def nuke_epub_cache():
+    """
+    Perform a sanity check first. The folder must exist and be named "epub".
+    """
+    epub_cache_path = cfg.epub_cache_path()
+    if not epub_cache_path.is_dir():
+        logger.info("Epub cache folder does not exist. Nothing to do.")
+        return
+
+    if epub_cache_path.name != "epubs":
+        logger.error(f"Epub cache folder is not named 'epubs', instead {epub_cache_path}. Aborting.")
+        return
+
+    try:
+        shutil.rmtree(epub_cache_path)
+    except OSError as e:
+        logger.error("Failed to delete epub cache folder.", exc_info=e)
+        return
